import pytoml as toml

from .log import log

class Config(object):
    def __init__(s, filename):
        s.config = {}
        s.load(filename)
        s.set_defaults()

    def set_default(s, key, val):
        global _config
        s.config[key] = s.config.get(key, val)

    def load(s, config_file):
        log.info("loading configuration file %s", config_file)
        try:
            with open(config_file, "r") as f:
                s.config = toml.loads(f.read())
        except FileNotFoundError as e:
            log.error("configuration file %s not found!" % config_file)
            raise e

    def set_defaults(s):
        pass

    def __getattr__(s, key):
        return s.config[key]

    def __str__(s):
        return str(s.config)



import logging

log = logging.getLogger('')
log.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

handler = logging.StreamHandler()
handler.setFormatter(formatter)

log.addHandler(handler)



import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.websocket
import json
import os

from threading import Lock

from .log import log
from .util import config

config.set_default("url_prefix", r"")

class GithubWebhook(object):
    def __init__(s, port, prs, github_handlers):

        s.secret = "__secret"
        s.port = port
        s.application = tornado.web.Application([
#            (r"/", GithubWebhook.MainHandler),
            (config.url_prefix + r"/api/pull_requests", GithubWebhook.PullRequestHandler, dict(prs=prs)),
            (config.url_prefix + r"/github", GithubWebhook.GithubWebhookHandler, dict(handler=github_handlers)),
            (config.url_prefix + r"/status", GithubWebhook.StatusWebSocket),
            (config.url_prefix + r"/control", GithubWebhook.ControlHandler),
                ])
        s.server = tornado.httpserver.HTTPServer(s.application)
        s.server.listen(s.port)
        s.websocket_lock = Lock()
        s.status_websockets = set()

    def run(s):
        log.info("tornado IOLoop started.")
        tornado.ioloop.IOLoop.instance().start()

    class MainHandler(tornado.web.RequestHandler):
        def get(self):
            self.write("...")

    class PullRequestHandler(tornado.web.RequestHandler):
        def initialize(s, prs):
            s.prs = prs

        def get(self):
            def gen_pull_entry(pr, job, time, extra = None):
                res = extra or {}
                res.update({
                        "title" : pr.title,
                        "user" : pr.user,
                        "url" : pr.url,
                        "commit" : job.arg,
                        "since" : time,
                        })
                return res

            self.set_header("Content-Type", 'application/json; charset="utf-8"')
            self.set_header("Access-Control-Allow-Credentials", "false")
            self.set_header("Access-Control-Allow-Origin", "*")

            building, queued, finished = self.prs.list()
            response = {}

            if building:
                _building = []
                for pr, job in building:
                    _building.append(
                            gen_pull_entry(pr, job, job.time_started))
                response['building'] = _building

            if queued:
                _queued = []
                for pr, job in queued:
                    _queued.append(
                            gen_pull_entry(pr, job, job.time_queued))

                response['queued'] = _queued

            if finished:
                _finished = []
                for pr, job in finished:
                    job_path_rel = os.path.join(pr.base_full_name, str(pr.nr), job.arg)
                    job_path_local = os.path.join(config.data_dir, job_path_rel)
                    job_path_url = os.path.join(config.http_root, job_path_rel)

                    extras = {
                        "output_url": os.path.join(job_path_url, "output.html"),
                        "result": job.result.name,
                        "runtime": (job.time_finished - job.time_started),
                      }
                    status_jsonfile = os.path.join(job_path_local,
                                                   "prstatus.json")
                    if os.path.isfile(status_jsonfile):
                        with open(status_jsonfile) as f:
                            # Content is up for interpretation between backend
                            # and frontend scripting
                            extras["status"] = json.load(f)
                    if "status" not in extras:
                        status_html_snipfile = os.path.join(
                                job_path_local, "prstatus.html.snip"
                            )

                        status_html = ""
                        if os.path.isfile(status_html_snipfile):
                            with open(status_html_snipfile, "r") as f:
                                status_html = f.read()
                        extras["status_html"] = status_html

                    _finished.append(
                            gen_pull_entry(pr, job, job.time_finished, extras)
                        )

                response['finished'] = _finished

            self.write(json.dumps(response, sort_keys=False, indent=4))

    class GithubWebhookHandler(tornado.web.RequestHandler):
        def initialize(s, handler):
            s.handler = handler

        def post(s):
            s.write("ok")
            hook_type = s.request.headers.get('X-Github-Event')

            handler = s.handler.get(hook_type)
            if handler:
                handler(s.request)
            else:
                log.warning("unhandled github event: %s", hook_type)

    class StatusWebSocket(tornado.websocket.WebSocketHandler):
        lock = Lock()
        websockets = set()
        keeper = None

        # passive read only websocket, so anyone can read
        def check_origin(self, origin):
            return True

        def write_message_all(message, binary=False):
            s = GithubWebhook.StatusWebSocket
            with s.lock:
                for websocket in s.websockets:
                    websocket.write_message(message, binary)

        def keep_alive():
            s = GithubWebhook.StatusWebSocket
            with s.lock:
                for websocket in s.websockets:
                    websocket.ping("ping".encode("ascii"))

        def open(self):
            print("websocket opened")
            with self.lock:
                if not self.websockets:
                    s = GithubWebhook.StatusWebSocket
                    s.keeper = tornado.ioloop.PeriodicCallback(s.keep_alive, 30*1000)
                    s.keeper.start()
                self.websockets.add(self)

        def on_message(self, message):
            pass

        def on_close(self):
            with self.lock:
                self.websockets.discard(self)
                if not self.websockets:
                    self.keeper.stop()

    class ControlHandler(tornado.web.RequestHandler):
        def post(self):
#            data = json.loads(self.request.body)
            s = GithubWebhook.StatusWebSocket
            s.write_message_all(self.request.body)



import os
import sys

from .config import Config

class MurdockConfig(Config):
    def set_defaults(s):
        super().set_defaults()
        s.config["fail_labels"] = set(s.config.get("fail_labels", []))
        s.set_default("ci_ready_label", "Ready for CI build")
        s.set_default("context", "Murdock")
        s.set_default("port", 3000)
        s.set_default("scripts_dir", os.getcwd() + "/scripts")
        s.set_default("set_status", False)
        s.set_default("sigterm_timeout", 100)
        s.set_default("github_username", None)
        s.set_default("github_password", None)
        s.set_default("github_apikey", None)

if len(sys.argv) > 1:
    config_file = sys.argv[1]
else:
    config_file = "/etc/murdock.toml"

config = MurdockConfig(config_file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['murdock_ci']

package_data = \
{'': ['*']}

install_requires = \
['agithub>=2.2.2,<3.0.0', 'tornado>=6.0.4,<7.0.0']

entry_points = \
{'console_scripts': ['murdock = murdock_ci:main']}

setup_kwargs = {
    'name': 'murdock-ci',
    'version': '0.1.3',
    'description': 'A simple gihub CI server',
    'long_description': None,
    'author': 'Kaspar Schleiser',
    'author_email': 'kaspar@schleiser.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
from enum import Enum
from threading import Lock
import time

from .log import log

class JobState(Enum):
    created = 0
    queued = 1
    running = 2
    finished = 3

class JobResult(Enum):
    passed = 0
    canceled = 1
    errored = 2
    timeout = 3
    unknown = 4

class Job(object):
    id = 0
    def __init__(s, name, cmd, env=None, hook=None, arg=None):
        s.lock = Lock()
        s.id = Job.id
        Job.id += 1

        s.name = name
        s.cmd = cmd
        s.env = env
        s.worker = None

        s.hook = hook
        s.arg = arg

        s.result = JobResult.unknown

        s.time_created = -1
        s.time_queued = -1
        s.time_started = -1
        s.time_finished = -1

        s.set_state(JobState.created)

    def data_dir(s):
        return s.name #,os.path.join(config.data_dir, s.name + "." + str(s.id))

    def set_state(s, state, result=JobResult.unknown):
        with s.lock:
            s.state = state
            s.result = result
            if s.state == JobState.created:
                s.time_created = time.time()
            elif s.state == JobState.queued:
                s.time_queued = time.time()
            elif s.state == JobState.running:
                s.time_started = time.time()
            elif s.state == JobState.finished:
                s.time_finished = time.time()

            log.info("Job %s: new state: %s %s %s %s %s %s " , s.name, s.state, s.result, s.time_created, s.time_queued, s.time_started, s.time_finished)

        if s.hook:
            s.hook(s.arg, s)

    def stopped(s, result):
        with s.lock:
            s.state = result

    def cancel(s):
        if s.worker:
            s.worker.cancel(s)
        else:
            s.set_state(JobState.finished, JobResult.canceled)



#!/usr/bin/env python3

import os
import subprocess
import json
import signal
import shutil
import tornado.ioloop
import traceback

import threading
from queue import Queue, Empty
from threading import Lock

from agithub.GitHub import GitHub

from .log import log
from .jobs import Job, JobResult, JobState
from .github_webhook import GithubWebhook
from .util import config


known_actions = { "labeled", "unlabeled", "synchronize", "created", "assigned",
        "closed", "edited", "unassigned", "opened", "status", "reopened", "review_requested" }


def nicetime(time):
    secs = round(time)
    minutes = secs/60
    hrs = minutes/60
    days = int(hrs/24)
    secs = int(secs % 60)
    minutes = int(minutes % 60)
    hrs = int(hrs % 24)
    res = ""
    if days:
        res += "%sd:" % days
    if hrs:
        res += "%sh:" % hrs
    if minutes:
        res += "%sm:" % minutes
    res += "%ss" % secs
    return res

class ShellWorker(threading.Thread):
    _lock = Lock()
    num_workers = 0

    def __init__(self, queue):
        threading.Thread.__init__(self, daemon=True)
        self.process = None
        self.queue = queue
        self.canceled = False
        self.job = None
        with ShellWorker._lock:
            ShellWorker.num_workers += 1
            self.num = ShellWorker.num_workers
        self.start()

    def run(s):
        log.info("ShellWorker %s: started.", s.num)
        while True:
            try:
                try:
                    s.job = None
                    s.process = None
                    s.canceled = False
                    job = s.queue.get()
                    s.job = job
                except Empty:
                    return
                if job.state == JobState.finished:
                    log.info("ShellWorker %s: skipping finished job %s", s.num, job.name)
                    s.queue.task_done()
                    continue
                else:
                    log.info("ShellWorker %s: building job %s", s.num, job.name)

                s.job = job
                job.worker = s
                s.job.set_state(JobState.running)
                s.job.env["CI_BUILD_ID"] = str(s.job.time_started)

                build_dir = os.path.join(s.job.data_dir(), "build")
                try:
                    os.makedirs(build_dir)
                except FileExistsError:
                    shutil.rmtree(build_dir)
                    os.makedirs(build_dir)

                _env = os.environ.copy()
                _env.update(s.job.env)

                output_file = open(os.path.join(s.job.data_dir(), "output.txt"), mode='wb')
                s.process = p = subprocess.Popen([ s.job.cmd, "build" ],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             cwd=s.job.data_dir(), env=_env, start_new_session=True)
                s.job.worker = s
                try:
                    for line in p.stdout:
                        output_file.write(line)
                except Exception as e:
                    log.info(e)
                finally:
                    output_file.close()

                log.info("ShellWorker %s: Job %s finished. result: %s", s.num, s.job.name, s.job.result)

                if s.process:
                    s.process.wait()

                try:
                    subprocess.check_call([s.job.cmd, "post_build"], cwd=s.job.data_dir(), env=_env)
                except subprocess.CalledProcessError:
                    log.warning("Job %s: post build script failed.", s.job.name)
                    pass

                if s.canceled:
                    s.job.set_state(JobState.finished, JobResult.canceled)
                else:
                    ret = s.process.returncode
                    s.process = None
                    if ret == 0:
                        s.job.set_state(JobState.finished, JobResult.passed)
                    else:
                        s.job.set_state(JobState.finished, JobResult.errored)

                try:
                    shutil.rmtree(build_dir)
                except FileNotFoundError:
                    pass

                s.queue.task_done()

            except Exception as e:
               log.warning("ShellWorker %s: uncaught exception: %s", s.num, e)
               traceback.print_exc()

    def cancel(s, job):
        if s.process is not None and s.job==job:
            threading.Thread(target=ShellWorker.graceful_kill, args=(s.process,)).start()
            s.process=None
            s.canceled=True

    def graceful_kill(process):
        try:
            os.killpg(process.pid, signal.SIGTERM)
            process.wait(config.sigterm_timeout)
        except subprocess.TimeoutExpired:
            log.warning("ShellWorker: killing process")
            process.kill()
            process.wait()

class PullRequest(object):
    _map = {}

    def __init__(s, data):
        s.data = data
        s._map[data["_links"]["html"]["href"]] = s
        s.current_job = None
        s.jobs = []
        s.labels = set()
        s.old_head = None

    def get(data, create=True):
        if "pull_request" in data:
            data = data["pull_request"]

        pull_url = data["_links"]["html"]["href"]
        pr = PullRequest._map.get(pull_url)
        if pr:
            pr.data = data
            log.info("PR %s updated", pr.url)
        else:
            if not create:
                return None

            pr = PullRequest(data)
            log.info("PR %s new to Murdock (state=%s, mergeable=%s, merge_commit_sha=%s)", pr.url, pr.state, pr.mergeable, pr.merge_commit)
            pr.update_labels()
        return pr

    def close(data):
        pr = PullRequest.get(data, False)
        if pr:
            pr.cancel_job()
            log.info("PR %s: closed.", pr.url)
        else:
            log.warning("PR %s unknown, but tried to close!", data["_links"]["html"]["href"])

    def update(s):
        if s.head != s.old_head:
            s.old_head = s.head
            log.info("PR %s has new head: %s", s.url, s.head)
            if (s.state=="open") \
                and config.ci_ready_label in s.labels \
                and s.mergeable in { True, None }:
                s.start_job()
        return s

    def cancel_job(s):
        if s.current_job and s.current_job.state!=JobState.finished:
            log.info("PR %s: canceling build of commit %s", s.url, s.current_job.arg)
            s.current_job.cancel()
            s.current_job = None
        return s

    def start_job(s):
        s.cancel_job()

        log.info("PR %s: queueing build of commit %s", s.url, s.head)

        env = { "CI_PULL_COMMIT" : s.head,
                "CI_PULL_REPO" : s.repo,
                "CI_PULL_BRANCH" : s.branch,
                "CI_PULL_NR" : str(s.nr),
                "CI_PULL_URL" : s.url,
                "CI_PULL_TITLE" : s.title,
                "CI_PULL_USER" : s.user,
                "CI_BASE_REPO" : s.base_repo,
                "CI_BASE_BRANCH" : s.base_branch,
                "CI_BASE_COMMIT" : s.base_commit,
                "CI_SCRIPTS_DIR" : config.scripts_dir,
                "CI_PULL_LABELS" : ";".join(sorted(list(s.labels))),
                "CI_BUILD_HTTP_ROOT" : os.path.join(config.http_root,
                                       s.base_full_name, str(s.nr), s.head),
                }

        if s.mergeable:
            env["CI_MERGE_COMMIT"] = s.merge_commit

        for key, value in env.items():
            if not value:
                log.warning("PR %s: env %s has NoneType!", s.url, key)
                return s

        s.current_job = Job(s.get_job_path(s.head), os.path.join(config.scripts_dir, "build.sh"), env, s.job_hook, s.head)
        s.jobs.append(s.current_job)
        queue.put(s.current_job)

        s.current_job.set_state(JobState.queued)
        return s

    def get_job_path(s, commit):
        return os.path.join(config.data_dir, s.base_full_name, str(s.nr), commit)

    def update_labels(s):
        code, result = github.repos[s.base_full_name].issues[s.nr].labels.get()
        if code == 200:
            s.labels = set()
            for label in result:
                log.info("PR %s set label: %s", s.url, label["name"])
                s.labels.add(label["name"])
        return s

    def add_label(s, label):
        log.info("PR %s added label: %s", s.url, label)
        if label in s.labels:
            log.warning("PR %s label already present.", s.url)
            return
        s.labels.add(label)
        if label == config.ci_ready_label:
            s.start_job()
        return s

    def remove_label(s, label):
        log.info("PR %s removed label: %s", s.url, label)
        s.labels.discard(label)
        if label == config.ci_ready_label:
            s.cancel_job()
        return s

    def __getattr__(s, field):
        if field == "url":
            return s.data["_links"]["html"]["href"]
        elif field == "base_repo":
            return s.data["base"]["repo"]["clone_url"]
        elif field == "base_branch":
            return s.data["base"]["ref"]
        elif field == "base_commit":
            return s.data["base"]["sha"]
        elif field == "base_full_name":
            return s.data["base"]["repo"]["full_name"]
        elif field == "nr":
            return s.data["number"]
        elif field == "branch":
            return s.data["head"]["ref"]
        elif field == "repo":
            return s.data["head"]["repo"]["clone_url"]
        elif field == "head":
            return s.data["head"]["sha"]
        elif field == "user":
            return s.data["head"]["user"]["login"]
        elif field == "title":
            return s.data["title"]
        elif field == "state":
            return s.data["state"]
        elif field == "merge_commit":
            return s.data["merge_commit_sha"]
        elif field == "mergeable":
            return s.data["mergeable"]
        else:
            raise AttributeError

    def job_hook(s, arg, job):
        target_url = None
        runtime = None
        if job.state == JobState.created:
            state = "pending"
            description = "The build has been queued."
        elif job.state == JobState.running:
            state = "pending"
            description = "The build has been started."
        elif job.state == JobState.finished:
            runtime = job.time_finished - job.time_started
            target_url = os.path.join(config.http_root, s.base_full_name, str(s.nr), arg, "output.html")
            if job.result == JobResult.passed:
                if not s.labels & config.fail_labels:
                    state = "success"
                    description = "The build succeeded. runtime: %s" % nicetime(runtime)
                else:
                    state = "error"
                    description = "The build only failed the label check. runtime: %s" % nicetime(runtime)
                    job.result = JobResult.errored
            elif job.result == JobResult.errored:
                state = "error"
                description = "The build failed. runtime: %s" % nicetime(runtime)
            else:
                state = "failure"
                target_url = None
                if job.result == JobResult.canceled:
                    description = "The build has been canceled."
                else:
                    description = "The build has failed for an unknown reason."
        else:
            return

        status = {
                "state": state,
                "description": description,
                }

        if target_url:
            status["target_url"] = target_url
        else:
            status["target_url"] = config.http_root

        s.set_status(arg, **status)

        if runtime:
            log.info("PR %s runtime: %s", s.url, nicetime(runtime))

        log.info("PR %s notifying webhooks", s.url)
        GithubWebhook.StatusWebSocket.write_message_all('{ "cmd" :"reload_prs" }')

    def set_status(s, commit, **kwargs):
        status = {
                "state" : "failure",
                "description" : "unknown reason",
                "context" : config.context
                }

        status.update(kwargs)

        if not config.set_status:
            log.info("PR %s not setting github status: %s \"%s\"", s.url, status["state"], status["description"])
            return

        log.info("PR %s setting github status: %s \"%s\"", s.url, status["state"], status["description"])
        github.repos[s.base_full_name].statuses[commit].post(body=status)

    def cancel_all():
        log.info("canceling jobs...")
        for url, pr in PullRequest._map.items():
            if pr.current_job:
                pr.current_job.cancel()

    def load(repo):
        code, result = github.repos[repo].pulls.get()
        if code==200:
            for data in result:
                if data["state"] == "open":
                    pr = PullRequest.get(data)
                    if pr.current_job:
                        continue
                    if not config.ci_ready_label in pr.labels:
                        continue
                    state = pr.get_state()
                    if state == "canceled" or state == "pending":
                        pr.start_job()

    def get_state(s):
        code, result = github.repos[s.base_full_name].statuses[s.head].get()
        if code==200:
            for data in result:
                if data["context"] == config.context:
                    if data["description"] == "The build has been canceled.":
                        return "canceled"
                    else:
                        return data["state"]
        else:
            log.warning("PullRequest: couldn't get statuses: code %s", code)

        return "unknown"

    def list():
        building = []
        queued = []
        finished = []
        for name, pr in PullRequest._map.items():
            job = pr.current_job
            if job:
                if job.state==JobState.finished:
                    finished.append((pr, job))
                elif job.state==JobState.running:
                    building.append((pr, job))
                else:
                    queued.append((pr, job))

        building = sorted(building, key=lambda x: x[1].time_started)
        queued = sorted(queued, key=lambda x: x[1].time_queued)
        finished = sorted(finished, key=lambda x: x[1].time_finished)

        return (building, queued, finished)

handle_pull_request_lock = Lock()

def handle_pull_request(request):
    with handle_pull_request_lock:
        data = json.loads(request.body.decode("utf-8"))
        pr_data = data["pull_request"]
        repo = pr_data["base"]["repo"]["full_name"]
        if not repo in config.repos:
            log.warning("ignoring PR for repo %s", repo)
            return

        #print(json.dumps(data, sort_keys=False, indent=4))
        action = data["action"]

        pr_url = pr_data["_links"]["html"]["href"]
        log.info("PR %s hook action %s", pr_url, action)

        if not action in known_actions:
            log.warning("PR %s unknown action %s", pr_url, action)
            log.debug(json.dumps(data, sort_keys=False, indent=4))

        if action in { "closed" }:
            PullRequest.close(pr_data)
            return

        pr = PullRequest.get(pr_data).update()
        if action == "unlabeled":
            pr.remove_label(data["label"]["name"])
        elif action == "labeled":
            pr.add_label(data["label"]["name"])
        elif action in { "created", "opened", "reopened" } and not config.ci_ready_label in pr.labels:
            status = {
                    "description": "\"%s\" label not set" % config.ci_ready_label,
                    "target_url" : config.http_root,
                    }

            pr.set_status(pr_data["head"]["sha"], **status)

def handle_push(request):
    data = json.loads(request.body.decode("utf-8"))

    log.info(json.dumps(data, sort_keys=False, indent=4))

github_handlers = {
        "pull_request" : handle_pull_request,
#        "push" : handle_push,
        }
if not (bool(config.github_username and config.github_password) ^ bool(config.github_apikey)):
    raise SystemExit("No valid github authentication provided, provide "
                     "username/password or an API key in the configuration "
                     "file.")
github = GitHub(config.github_username, config.github_password, token=config.github_apikey)
queue = Queue()
ShellWorker(queue)

def shutdown():
    log.info("murdock: shutting down.")
    tornado.ioloop.IOLoop.instance().stop()

def sig_handler(sig, frame):
    log.warning('Caught signal: %s', sig)
    shutdown()

def startup_load_pull_requests():
    log.info("Loading pull requests...")
    for repo in config.repos:
        PullRequest.load(repo)
    log.info("All pull request loaded.")

def main():
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    log.info("murdock initialized.")

#    threading.Thread(target=startup_load_pull_requests, daemon=True).start()

    g = GithubWebhook(config.port, PullRequest, github_handlers)
    g.run()

    # tornado loop ended

    PullRequest.cancel_all()
    log.info("murdock shut down.")
